#ifndef BFM_CG_H
#define BFM_CG_H

/*  2007
 *  Calls PAB's assembler routines to give an implementation
 *  of the dwf dslash. Aim is to give very high performance
 *  in a reasonably portable/retargettable library.
 */
#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>
#ifdef PPC440QCDOC
#include <timersub.h>
#include <qcdocos/scu_checksum.h>
#endif
#include <sys/time.h>
template <class Float>
int bfmbase<Float>::mprecFlopsPerSite ( void ) 
{
  int fourdwilson = 1320;
  int fivedhop    = 24;

  int flops;

  if ( precon_5d && (Ls > 1) ) {
    // A + B d d
    flops =  2*(fourdwilson + fivedhop) + 72;
  } else if ( Ls > 1 ) {
    flops = 2*(fourdwilson + 24*5);
  } else {
    // A+ B d d 
    flops =  2*fourdwilson + 72;
  }
  return flops;
}

#ifdef PPC440QCDOC
typedef double fenv_t;
typedef union
{
  fenv_t fenv;
  unsigned int l[2];
} fenv_union_t;
#define fegetenv_register() \
        ({ fenv_t env; asm volatile ("mffs %0" : "=f" (env)); env; })

#define fesetenv_register(env) \
        ({ double d = (env); asm volatile ("mtfsf 0xff,%0" : : "f" (d)); })

#endif


template <class Float>
int bfmbase<Float>::CGNE(Fermion_t solution[2], Fermion_t source[2])
{

#ifdef PPC440QCDOC
  if ( !ScuChecksum::ChecksumsOn())
    ScuChecksum::Initialise(true,true);

  fenv_union_t nan;
  nan.fenv = fegetenv_register();
  nan.l[1] = nan.l[1] | 0xF0;
  //  fesetenv_register(nan.fenv);

  if ( ! ScuChecksum::CsumSwap() ){
    QDP_error_exit("SCU checksum mismatch before CG");
  }
#endif

  Fermion_t src = allocFermion(); 
  Fermion_t tmp = allocFermion(); 
  Fermion_t Mtmp= allocFermion(); 

  // src_o = source_o - Moe MeeInv source_e
  double f;
  f = norm(source[0]);
  f+= norm(source[1]);
  if ( isBoss() ) printf("Source norm is %le\n",f);

  f = norm(solution[0]);
  f+= norm(solution[1]);
  if ( isBoss() ) printf("Guess norm is %le\n",f);

  MooeeInv(source[0],tmp,0);
  dslash  (tmp,src,1,0);
  axpy    (tmp,src,source[1],0.5);

  Mprec(tmp,src,Mtmp,1);  

  freeFermion(Mtmp);

  int iter;

  if ( reproduce ) { 

    Fermion_t tmp_r = allocFermion(); 
    axpy    (tmp_r,solution[1],solution[1],0.0);
    repro_control(ReproRecord);
    iter = CGNE_prec(tmp_r,src);

    repro_control(ReproCheck);
    iter = CGNE_prec(solution[1],src);

    repro_control(ReproIdle);
    freeFermion(tmp_r);

  } else {

    repro_control(ReproIdle);
    iter = CGNE_prec(solution[1],src);

  }

  // sol_e = M_ee^-1 * ( src_e - Meo sol_o )...
  dslash(solution[1],tmp,0,0);
  axpy(src,tmp,source[0],0.5);
  MooeeInv(src,solution[0],0);

  freeFermion(tmp);
  freeFermion(src);

#ifdef PPC440QCDOC
  if ( ! ScuChecksum::CsumSwap() ){
    QDP_error_exit("SCU checksum mismatch before CG");
  }
#endif

  return iter;
}

template <class Float>
int bfmbase<Float>::CGNE_prec(Fermion_t psi, Fermion_t src)
{
  Fermion_t p   = allocFermion(mem_fast); 
  Fermion_t tmp = allocFermion(mem_fast); 
  Fermion_t mp  = allocFermion(mem_fast); 
  Fermion_t mmp = allocFermion(mem_fast); 
  Fermion_t r   = allocFermion(mem_fast); 

  double f;
  double cp,c,a,d,b;

  if ( isBoss() ) printf ("CGNE_prec\n");
  
  //Initial residual computation & set up
  Mprec(psi,mp,tmp,0);
  Mprec(mp,mmp,tmp,1);

  axpy (r, mmp, src,-1.0);
  axpy (p, r, r,0.0);

  cp= norm(r);
  Float ssq = norm(src);
  Float rsq =  residual* residual*ssq;

  //Check if guess is really REALLY good :)
  if ( cp <= rsq ) {
    if ( isBoss() ) printf("bfmbase::CGNE_prec k=0 converged - nice guess\n");
    freeFermion(tmp);
    freeFermion(p);
    freeFermion(mp);
    freeFermion(mmp);
    freeFermion(r);
    return 0;
  }
  if ( isBoss() ) printf("bfmbase::CGNE_prec k=0 residual %le rsq %le\n",cp,rsq);

  //  if ( isBoss() ) {
  if ( 0 ) {
    printf("CG: psi %8.8x \n",psi);
    printf("CG: tmp %8.8x \n",tmp);
    printf("CG: p   %8.8x \n",p);
    printf("CG: mp  %8.8x \n",mp);
    printf("CG: mmp %8.8x \n",mmp);
    printf("CG: r   %8.8x \n",r);
    printf("CG: u   %8.8x \n",u);
  }
  struct timeval start,stop;
  gettimeofday(&start,NULL);
  for (int k=1;k<=max_iter;k++){

    c=cp;

    // Mp
    d = Mprec(p,mp,tmp,0,1);

    // Update solution vector
    a = c/d;
    axpy(psi,p,psi,a);

    // MdagMp
    Mprec(mp,mmp,tmp,1); 

    //Fused axpy-norm required.    
    //As know "a" in advance could save bandwidth by
    //not writing out mmp, but doing linear combination.
    //in Mprec.
    cp = axpy_norm(r,mmp,r,-a);

    if ( k%100 == 0 )
      if ( isBoss() ) printf("bfmbase::CGNE_prec: k=%d r^2=%le %le\n",k,cp,sqrt(cp/ssq));

    // Stopping condition
    if ( cp <= rsq ) { 

      gettimeofday(&stop,NULL);
      struct timeval diff;
      timersub(&stop,&start,&diff);

      if ( isBoss() ) printf("bfmbase::CGNE_prec converged in %d iterations\n",k);
      if ( isBoss() ) printf("bfmbase::CGNE_prec converged in %d.%6.6d s\n",diff.tv_sec,diff.tv_usec);

      int lx = node_latt[0];
      int ly = node_latt[1];
      int lz = node_latt[2];
      int lt = node_latt[3];
      int sites = (lx*ly*lz*lt*Ls)/2;

      double flops = 1.0*sites*k*(  2.0*mprecFlopsPerSite() // Two matrix multiples
				    + 48*3                  // Three AXPY's
				    + 48*2  );              // Two   Norms (one in Mprec)

      double t = diff.tv_sec*1.0E6 + diff.tv_usec;
      if ( isBoss() ) printf("%d mprec flops/site\n",mprecFlopsPerSite());
      if ( isBoss() ) printf("%le flops\n",flops);
      if ( isBoss() ) printf("%le mflops\n",flops/t);

      Mprec(psi,mp,tmp,0);
      Mprec(mp,mmp,tmp,1); 
      axpy(tmp,src,mmp,-1.0);
      double true_residual = sqrt(norm(tmp)/norm(src));
      if ( isBoss() ) printf("true residual is %le \n",true_residual);
      Float ssq = norm(src);

      freeFermion(tmp);
      freeFermion(p);
      freeFermion(mp);
      freeFermion(mmp);
      freeFermion(r);
      return k;
    }
    // New (conjugate/M-orthogonal) search direction
    b = cp/c;
    axpy(p,p,r,b);

  }
  if ( isBoss() ) printf("CG not converged \n");
  freeFermion(tmp);
  freeFermion(p);
  freeFermion(mp);
  freeFermion(mmp);
  freeFermion(r);
  return -1;

}
#endif

