#ifndef BFM_MPREC_H
#define BFM_MPREC_H
/*  2007
 *  Calls PAB's assembler routines to give an implementation
 *  of the dwf dslash. Aim is to give very high performance
 *  in a reasonably portable/retargettable library.
 */
#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>


template <class Float>
void bfmbase<Float>::dump(Fermion_t psi, char *filename, double scale, int cb)
{
  FILE *fp = fopen(filename,"w");
  Float *ptr =  (Float *)psi;

  int x[4];
  for(int s=0;s<this->Ls;s++){
  for(x[3]=0;x[3]<this->node_latt[3];x[3]++){
  for(x[2]=0;x[2]<this->node_latt[2];x[2]++){
  for(x[1]=0;x[1]<this->node_latt[1];x[1]++){
  for(x[0]=0;x[0]<this->node_latt[0];x[0]++){
    int reim = 0;
    int co=0;
    if ( ((x[3]+x[2]+x[1]+x[0]+s)&0x1) == cb ) { 
              
      int base = bagel_idx5d (x,s,reim,co,12,1);
      for(int cpt=0;cpt<24;cpt++){
	fprintf(fp,"%d %d %d %d %d %d : %le\n",x[0],x[1],x[2],x[3],s,cpt,ptr[base+cpt]*scale);
      }
    }
  }}}}
  }
  fclose(fp);
}

static int dumped;

template <class Float>
double bfmbase<Float>::Mprec(Fermion_t psi, 
			     Fermion_t chi, 
			     Fermion_t tmp, 
			     int dag,
			     int donrm)
{

  double nrm = 0.0;

  if ( (!precon_5d) && (Ls > 1) ) { 

    dslash  (psi,tmp,0,dag);   // odd->even -- implicit -0.5 factor to be applied
    MooeeInv(tmp,chi,dag);     // even->even * 1/(5-M5)
    dslash  (chi,tmp,1,dag);   // even->odd -- implicit -0.5 factor to be applied
    Mooee   (psi,chi,dag);     // odd odd * (5-M5)

    double m1=-0.25;
    nrm = axpy_norm (chi,tmp,chi,m1);     // Linear combination

  } else {

    double ident = (5.-M5);
    double hop   = -0.25/(ident);

    dslash  (psi,chi,0,dag);   // odd->even -- implicit -0.5 factor to be applied
    if ( reproduce ) nrm = norm(chi) ;
    dslash  (chi,tmp,1,dag);   // even->odd -- implicit -0.5 factor to be applied
    if ( reproduce ) nrm = norm(tmp) ;

    if ( donrm ) {
      nrm = axpby_norm(chi,tmp,psi,hop,ident);     // Linear combination
    }  else {
      axpby(chi,tmp,psi,hop,ident);
    }

  }

  return nrm;
}
template <class Float>
void bfmbase<Float>::Munprec(Fermion_t psi[2], 
			      Fermion_t chi[2], 
			      Fermion_t tmp,
		       int dag) 
{
  double m=-0.5;

  Mooee(psi[0],chi[0],dag);
  Mooee(psi[1],chi[1],dag);
  printf("Mooee[0] %le\n",this->norm(chi[0]));
  printf("Mooee[1] %le\n",this->norm(chi[1]));

  dslash(psi[0],tmp,1,dag);
  axpy  (chi[1],tmp,chi[1],m);     // Linear combination

  dslash(psi[1],tmp,0,dag);
  axpy  (chi[0],tmp,chi[0],m);     // Linear combination

  printf("axpy[0] %le\n",this->norm(chi[0]));
  printf("axpy[1] %le\n",this->norm(chi[1]));

}


// CPS style .. these are site diagonal and trivial
template <class Float>
void bfmbase<Float>::Mooee5dprec(Fermion_t psi_t, 
				  Fermion_t chi_t, 
				  int dag)
{
  Float *psi = (Float *) psi_t;
  Float *chi = (Float *) chi_t;
  int nspinco = 12;
  Float invtwokappa = 5.-M5;

  Float *xx;
  Float *rr;
  for(int site=0;site<simd_cbvol;site++){
   for(int s=0;s<cbLs;s++) {

    xx = &psi[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*cbLs];
    rr = &chi[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*cbLs];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int simd_el=0;simd_el<nsimd;simd_el++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + simd_el*2 + spinco*nsimd*2;
      rr[idx] = invtwokappa*xx[idx];
    }}}
   }
  }  
  return;
}
template <class Float>
void bfmbase<Float>::MooeeInv5dprec(Fermion_t psi_t, 
				     Fermion_t chi_t, 
				     int dag)
{
  Float *psi = (Float *) psi_t;
  Float *chi = (Float *) chi_t;
  int nspinco = 12;
  Float invtwokappa = 5.-M5;
  Float twokappa = 1.0/invtwokappa;

  Float *xx;
  Float *rr;
  for(int site=0;site<simd_cbvol;site++){
   for(int s=0;s<cbLs;s++) {

    xx = &psi[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*cbLs];
    rr = &chi[s*nspinco*nsimd*2   + site*nspinco*nsimd*2*cbLs];

    for(int spinco=0;spinco<nspinco;spinco++){
    for(int simd_el=0;simd_el<nsimd;simd_el++){
    for(int reim=0;reim<2;reim++){
      int idx = reim + simd_el*2 + spinco*nsimd*2;
      rr[idx] = twokappa*xx[idx];
    }}}
   }
  }  
  return;
}



// CHROMA style oo/ee part of preconditioned matrix
template <class Float>
void bfmbase<Float>::Mooee4dprec(Fermion_t psi_t, 
				  Fermion_t chi_t, 
				  int dag)
{
  Float *psi = (Float *) psi_t;
  Float *chi = (Float *) chi_t;
  int nspinco = 12;

  for(int site=0;site<simd_cbvol;site++){
  for(int s=0;s<Ls;s++) {

    Float invtwokappa = 5.-M5;

    Float  c,  c_p,  c_m ;
    Float *p, *p_p, *p_m, *ch ;
    integer    s_p,  s_m ;

    c_m = c_p = -1.0;
    c   = invtwokappa;

    if ( s== 0   ) c_m = mass; // mf boundary wrap
    if ( s== Ls-1 ) c_p = mass;

    s_p = (s+1)%Ls;
    s_m = (s+Ls-1)%Ls;

    // Real/Imaginary innermost
    // Can I use the virtual bagel_idx function????

    p   = &psi[bagel_idx5d(site,s,nspinco)];
    p_m = &psi[bagel_idx5d(site,s_m,nspinco)];
    p_p = &psi[bagel_idx5d(site,s_p,nspinco)];
    ch  = &chi[bagel_idx5d(site,s,nspinco)];
    /*
     * For bluegene this can be made very efficient
     * by each time around the s-loop loading 
     * L 6-regs psi   (+6 regs reusing last psi_p)
     * L 6-regs psi_p, 
     *   psi_m -> old psi regs 
     *  6/12-regs chi are assigned and stored.
     * Footprint is 24-30 registers.
     *
     * Need to unroll by two to avoid register copies.
     * Modulo scheduling would have been nice...
     * But unrolling & icaches are the reason modulo scheduling 
     * is always avoidable I guess....
     */

    for(int spinco=0;spinco<nspinco/2;spinco++){
    for(int simd_el=0;simd_el<nsimd;simd_el++){
    for(int reim=0;reim<2;reim++){

      int idx = reim + simd_el*2 + spinco*nsimd*2;
      int hdx = reim + simd_el*2 + (spinco+6)*nsimd*2;

      if ( dag ) {
        ch[idx]= c*p[idx]+c_p*p_p[idx];
        ch[hdx]= c*p[hdx]+c_m*p_m[hdx];
      }else{
        ch[idx]= c*p[idx]+c_m*p_m[idx];
        ch[hdx]= c*p[hdx]+c_p*p_p[hdx];
      }
    }}}

    
  }
  }
}

template <class Float>
void bfmbase<Float>::MooeeInv4dprec(Fermion_t psi_t, 
				     Fermion_t chi_t, 
				     int dag)
{
  Float *psi = (Float *) psi_t;
  Float *chi = (Float *) chi_t;

  int nspinco = 12;
  int idx;
  int hdx;
  int adx;

  Float invtwokappa = 5.-M5;
  Float twokappa = 1.0/(5.-M5);
  Float kappa    = 0.5*twokappa;
  Float invDfactor = 1.0/(1.0+mass/pow(invtwokappa,Ls));

  Float *p,*ch,*chm,*chp;

  for(int site=0;site<simd_cbvol;site++){

   Float ChiAdjust[12*simd()];
   bzero((void *)ChiAdjust,12*simd()*sizeof(Float));

   Float fact=mass*twokappa;

   for(int s=0;s<Ls;s++) {

    p   = &psi[bagel_idx5d(site,s,nspinco)];
    ch  = &chi[bagel_idx5d(site,s,nspinco)];

    for(int spinco=0;spinco<nspinco/2;spinco++){
    for(int simd_el=0;simd_el<simd();simd_el++){
    for(int reim=0;reim<2;reim++){

      if ( dag ) {
        hdx = reim + simd_el*2 + spinco*simd()*2;
        idx = reim + simd_el*2 + (spinco+6)*simd()*2;
	adx = hdx;
      } else {  
        idx = reim + simd_el*2 + spinco*simd()*2;
        hdx = reim + simd_el*2 + (spinco+6)*simd()*2;
        adx = idx;
      }

      // Copy and scale by two kappa
      ch[idx] = p[idx]*twokappa;
      ch[hdx] = p[hdx]*twokappa;

      // inverse of Lm
      if ( s<(Ls-1) ) {
	ChiAdjust[adx] -= fact*ch[hdx]; // P- takes lower components
      } else { 
	ch[hdx] = ch[hdx]+ChiAdjust[adx];
      }

      //inverse of L -- forward elimination
      if ( s>0 ) { 
        chm = &chi[bagel_idx5d(site,s-1,nspinco)];
	ch[idx]+=twokappa*chm[idx];
      }

      //Inverse of D
      if ( s==(Ls-1) ) { 
	ch[idx] *= invDfactor;
	ch[hdx] *= invDfactor;
      }

    }}}    

    fact*=twokappa;

   }

   fact=mass*pow(twokappa,Ls-1);
   // mass*pow(twokappa,s+1)
   for(int s=Ls-2;s>=0;s--) {
     Float *chl;
     ch  = &chi[bagel_idx5d(site,s,nspinco)];
     chp = &chi[bagel_idx5d(site,s+1,nspinco)];
     chl = &chi[bagel_idx5d(site,Ls-1,nspinco)];

     for(int spinco=0;spinco<nspinco/2;spinco++){
     for(int simd_el=0;simd_el<simd();simd_el++){
     for(int reim=0;reim<2;reim++){

       if ( dag ) {
	 hdx = reim + simd_el*2 + spinco*simd()*2;
	 idx = reim + simd_el*2 + (spinco+6)*simd()*2;
       } else {  
	 idx = reim + simd_el*2 + spinco*simd()*2;
	 hdx = reim + simd_el*2 + (spinco+6)*simd()*2;
       }

       ch[hdx] = ch[hdx]+twokappa*chp[hdx];
       ch[idx] -= fact*chl[idx];
     }}}
     fact /= twokappa;     
   }

  }
}
template <class Float>
void bfmbase<Float>::MooeeInv(Fermion_t psi, 
			       Fermion_t chi, 
			       int dag)
{
  if ( precon_5d ) 
    MooeeInv5dprec(psi,chi,dag);
  else
    MooeeInv4dprec(psi,chi,dag);
}
template <class Float>
void bfmbase<Float>::Mooee(Fermion_t psi, 
	             Fermion_t chi, 
	             int dag)
{
  if ( precon_5d ) 
    Mooee5dprec(psi,chi,dag);
  else
    Mooee4dprec(psi,chi,dag);
}

#endif 
