/****************************************************************************/
/* PAB Jan 2008                                                             */
/*
  // Methods implemented here
  virtual void init(bfmarg &arg) ; 
  virtual void end (void) ;

  virtual integer bagel_idx   (int site[4],int reim,int i, int i_size,int half_cb) ;
  virtual integer bagel_idx5d (int site[4],int s,int reim,int i, int i_size,int half_cb);
  virtual integer bagel_idx5d (int site,int s,int i_size) ;

  virtual void dslash(Fermion_t chi, 
		      Fermion_t psi, 
		      int cb,
		      int dag,int dperp=1) ;


  virtual void Dperp(Fermion_t psi, 
		     Fermion_t chi, 
		     int cb,
		     int dag);

  virtual int simd(void) { return 1;};
  virtual void pointers_init(void) ;
  virtual void pointers_end(void) ; 

  int  parity(int x[4]);
  int  psite (int x[4],int latt[4]);
  int  interleave_site(int pm,int mu, int site);

*/
/****************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "bfm.h"
#include <bfmbagel1.h>

#if defined(BFM_BAGEL1_BG) || defined(BFM_NOARCHBGL)
int BGL= 1; 
#else 
int BGL= 0;
#endif
template <class Float>
void bfmbagel1<Float>::init(bfmarg &arg)  
{
  int words;             /* size of the spinor field on the         */
  	  	  	 /* sublattice checkerboard                 */

  int slx;                      /* x-direction size of node sublattice */
  int sly;                      /* y-direction size of node sublattice */
  int slz;                      /* z-direction size of node sublattice */
  int slt;                      /* t-direction size of node sublattice */
  int i;
  int mu;

/*--------------------------------------------------------------------------*/
/* Set sublattice direction sizes                                           */
/*--------------------------------------------------------------------------*/

  * ((bfmarg *)this) = arg;

  // Hardwire 5d rb preconditioning for Bagel 1
  if ( this->precon_5d != 1 && this->Ls != 1) { 
    if ( this->isBoss() ) printf("bfmbagel1<Float>::init precon_5d logic bomb\n");
    exit(-1);
  }
  this->ncb=2;
  this->cbLs=this->Ls;
  this->node_cbvol = this->node_latt[0]
                    *this->node_latt[1]
                    *this->node_latt[2]
                    *this->node_latt[3]
                    /this->ncb;

  if ( this->isBoss() ) printf("bfmbagel1: nsimd =%d BAGEL_NSIMD=%d\n",simd(),BAGEL_NSIMD);
  this->nsimd = this->simd();
  if ( this->nsimd != 1 ) {
    if ( this->isBoss() ) printf("bfm<Float>::init nsimd logic bomb\n");
    exit(-1);
  }

  this->simd_grid[0] = 1;
  this->simd_grid[1] = 1;
  this->simd_grid[2] = 1;
  this->simd_grid[3] = 1;
    
  this->simd_latt[0] = slx = this->node_latt[0]/this->simd_grid[0];
  this->simd_latt[1] = sly = this->node_latt[1]/this->simd_grid[1];
  this->simd_latt[2] = slz = this->node_latt[2]/this->simd_grid[2];
  this->simd_latt[3] = slt = this->node_latt[3]/this->simd_grid[3];

/*--------------------------------------------------------------------------*/
/* Set periodic wrap back or not                                            */
/*--------------------------------------------------------------------------*/
  this->local_comm[0] = arg.local_comm[0];
  this->local_comm[1] = arg.local_comm[1];
  this->local_comm[2] = arg.local_comm[2];
  this->local_comm[3] = arg.local_comm[3];

/*-----------------------------------------------------------------------*/
/* compute the subgrd volume of each chkbd ... at least two local dims   */
/* must be even for this code to be correct.                             */
/*-----------------------------------------------------------------------*/
  this->simd_cbvol = (slx * sly * slz * slt)/this->ncb;
  
  this->simd_nbound[0] = (sly * slz * slt)/this->ncb; 
  this->simd_nbound[1] = (slx * slz * slt)/this->ncb;
  this->simd_nbound[2] = (slx * sly * slt)/this->ncb;
  this->simd_nbound[3] = (slx * sly * slz)/this->ncb;

  this->simd_allbound  = this->simd_nbound[0]
                       + this->simd_nbound[1]
                       + this->simd_nbound[2]
                       + this->simd_nbound[3];


  if ( this->simd_nbound[0] * slx * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even x logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[1] * sly * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even y logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[2] * slz * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even z logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[3] * slt * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even t logic bomb\n");
    exit(-1);
  }

  /*------------------------------------------------------------------------*/
  /* Check shape                                                            */
  /*------------------------------------------------------------------------*/
  if ( slx&1 ) {
    if ( this->isBoss() ) printf("Bagel is refusing to run as a sub latt is odd\n");
    exit(-1);
  }
  if ( (sly&1) &&(slz&1)&&(slt&1)  ) {
    if ( this->isBoss() ) printf("Bagel is refusing to run as y,z,t sub latts are all odd\n");
    exit(-1);
  }

  if ( BGL ) PAD_HALF_SPINOR_SIZE = 12;
  else  PAD_HALF_SPINOR_SIZE = 16;

  /*--------------------------------------------------------------------------*/
  /* Allocate two spinor temporaries                                          */
  /*--------------------------------------------------------------------------*/
  words = NMinusPlus * ND * PAD_HALF_SPINOR_SIZE * this->simd_cbvol;
  this->two_spinor[0] = (Float *)bfm_alloc(words*sizeof(Float),mem_fast); 
  this->two_spinor[1] = (Float *)bfm_alloc(words*sizeof(Float),mem_fast);
  if ( (this->two_spinor[0] == NULL) || (this->two_spinor[0] == NULL) ) {
     if ( this->isBoss() ) printf("BAGEL : warning two spinor alloc failed\n");
     exit(-1);
  }

  if ( this->isBoss() ) {
    printf("Two spinor arrays are 0x%lx 0x%lx\n",this->two_spinor[0],this->two_spinor[1]);
  }
/*--------------------------------------------------------------------------*/
/* Reserve memory for gauge field                                           */
/*--------------------------------------------------------------------------*/
  words = GAUGE_SIZE * this->nsimd * this->simd_cbvol *2*2;

  this->u = (Float *)bfm_alloc(words*sizeof(Float),mem_fast);
  if(this->u == 0){
    if ( this->isBoss() ) printf("bfm::u allocate\n");
    exit(-1);
  }
  if ( this->isBoss() ) printf("bfm::u 0x%8.8x\n",this->u);

/*--------------------------------------------------------------------------*/
/* Send and receive buffers                                                 */
/*--------------------------------------------------------------------------*/
  for(int pm=0;pm<2;pm++){
    for (int mu=0;mu<4;mu++){
      if ( !this->local_comm[mu] ){

      int bytes = this->simd_nbound[mu]*PAD_HALF_SPINOR_SIZE*sizeof(Float);
      
      for(int cb=0;cb<2;cb++){
	this->send_bufs[cb][pm][mu] = (Float *)bfm_alloc(bytes,mem_sendbuf);
	this->recv_bufs[cb][pm][mu] = (Float *)bfm_alloc(bytes,mem_fast);

	//	if ( this->isBoss() ) printf("bfm::bufs 0x%8.8x 0x%8.8x\n",this->send_bufs[cb][pm][mu],this->recv_bufs[cb][pm][mu]);

      }

      }
    }
  }

/*----------------------------------------------------------------------*/
/* Build the pointer table                                              */
/*----------------------------------------------------------------------*/
  this->pointers_init();
  
/*----------------------------------------------------------------------*/
/* Initialise the comms                                                 */
/*----------------------------------------------------------------------*/
  this->comm_init();

}
template <class Float>
void bfmbagel1<Float>::end (void)
{
  this->comm_end();
  this->pointers_end();
  bfm_free(this->two_spinor[0]);
  bfm_free(this->two_spinor[1]);
  bfm_free(this->u);

  for(int pm=0;pm<2;pm++){
  for(int cb=0;cb<2;cb++){
    for (int mu=0;mu<4;mu++){
      if ( !this->local_comm[mu] ){
	bfm_free(this->recv_bufs[cb][pm][mu]);
	bfm_free(this->send_bufs[cb][pm][mu]);
      }
    }
  }}
}

template <class Float>
integer bfmbagel1<Float>::bagel_idx5d (int site,int s,int nspinco) 
{
  return s*nspinco*this->simd_cbvol*this->simd()*2 + site*nspinco*this->simd()*2;
}
template <class Float>
integer bfmbagel1<Float>::bagel_idx5d (int x[4],int s,int reim,int i, int i_size, int half_cb)
{
  if ( !half_cb ) { 
    printf("Warning : bagel_idx called on both parities\n");
    printf("Warning : bagel_idx doesn't work for 5d vectors\n");
  }
  return s*i_size*this->simd_cbvol*this->simd()*2 + this->bagel_idx(x,reim,i,i_size,half_cb);
}
template <class Float>
integer bfmbagel1<Float>::bagel_idx (int x[4],int reim,int i, int i_size, int half_cb)
{
  int gx[4];
  int mu;

  for (mu=0;mu<4;mu++ ) gx[mu] = x[mu]/this->simd_latt[mu];

  integer simd_el = gx[0]+this->simd_grid[0]*( gx[1]
                               +this->simd_grid[1]*( gx[2]
                               +this->simd_grid[2]*  gx[3] ));
  int scoor[4]; // SIMD coordinate

  scoor[0] = x[0] % this->simd_latt[0]; /*Work out our location within the SIMD vol*/
  scoor[1] = x[1] % this->simd_latt[1];
  scoor[2] = x[2] % this->simd_latt[2];
  scoor[3] = x[3] % this->simd_latt[3];

  int scb   = this->parity(scoor);          /*Work out local checkerboard of site*/
  int ssite = this->psite(scoor,this->simd_latt); /*Work out checkerboarded site number*/

  // Real/Imaginary innermost
  int idx = reim + simd_el*2
                 + i*simd()*2
      + ssite*i_size*simd()*2;

// Bagel only deals with half checkerboard fermion vectors
// to allow for red black solution

  if ( !half_cb ) { 
    idx+= scb*this->simd()*2*i_size*this->simd_cbvol;
  }

  return idx;
}

static double complex_i[4] __attribute__ ((aligned(16))) = { 0.0, 1.0,0.0,1.0 } ;

extern "C" {
  void dec_hsu3    (double *psi,double *gauge,integer *len,integer *tab,double *ci);
  void dec_hsu3_dag(double *psi,double *gauge,integer *len,integer *tab,double *ci);

  void rec_su3    (double *psi,double *gauge,double *chiin,integer *len,double *ci);
  void rec_su3_dag(double *psi,double *gauge,double *chiin,integer *len,double *ci);

  void dec_hsu3_single    (float *psi,float *gauge,integer *len,integer *tab,double *ci);
  void dec_hsu3_dag_single(float *psi,float *gauge,integer *len,integer *tab,double *ci);

  void rec_su3_single    (float *psi,float *gauge,float *chiin,integer *len,double *ci);
  void rec_su3_dag_single(float *psi,float *gauge,float *chiin,integer *len,double *ci);
}


template <class Float>
void bfmbagel1<Float>::decom(Float *psi,Float *gauge,integer *len,integer *tab,int dag)
{
    if (sizeof(Float) == sizeof(double)) {
      if ( dag ) dec_hsu3_dag((double *)psi,(double *)gauge,len,tab,complex_i);
      else       dec_hsu3    ((double *)psi,(double *)gauge,len,tab,complex_i);
    } else {
      if ( dag ) dec_hsu3_dag_single((float *)psi,(float *)gauge,len,tab,complex_i);
      else       dec_hsu3_single    ((float *)psi,(float *)gauge,len,tab,complex_i);
    }
}
template <class Float>
void bfmbagel1<Float>::recon(Float *psi,Float *gauge,Float *chiin,integer *len,int dag)
{
    if (sizeof(Float) == sizeof(double)) {
      if ( dag ) rec_su3_dag((double *)psi,(double *)gauge,(double *)chiin,len,complex_i);
      else  	 rec_su3    ((double *)psi,(double *)gauge,(double *)chiin,len,complex_i);
    } else { 
      if ( dag ) rec_su3_dag_single((float *)psi,(float *)gauge,(float *)chiin,len,complex_i);
      else  	 rec_su3_single    ((float *)psi,(float *)gauge,(float *)chiin,len,complex_i);
    }
}


template <class Float>
void bfmbagel1<Float>::dslash(Fermion_t psi, // Input
                       	       Fermion_t chi, // Result
                               int cb,
                               int dag,int dperp)
{
  /*Gauge args*/
  Float *gauge[2];
  gauge[0] = (Float *)this->u ;
  gauge[1] = (Float *)this->u + GAUGE_SIZE*this->simd_cbvol;

// Loop over s applying Dw

  Float *psi_f = (Float *)psi;
  Float *chi_f = (Float *)chi;

  for(int s = 0; s< this->Ls; s+=2){

    int scb0=cb;
    int scb1=1-scb0;

    int offst0 = this->bagel_idx5d (0,s+0,12);
    Float *psi_s0 = &psi_f[offst0];
    Float *chi_s0 = &chi_f[offst0];

    int offst1 = this->bagel_idx5d (0,s+1,12);
    Float *psi_s1 = &psi_f[offst1];
    Float *chi_s1 = &chi_f[offst1];

    decom(psi_s0,gauge[1-scb0],&this->simd_cbvol,this->shift_table[scb0],dag);

    comm_start(scb0,psi_s0,dag); 
    decom(psi_s1,gauge[1-scb1],&this->simd_cbvol,this->shift_table[scb1],dag);
    comm_complete(scb0,psi_s0); 

    comm_start(scb1,psi_s1,dag); 
    recon(chi_s0,gauge[scb0],this->two_spinor[scb0],&this->simd_cbvol,dag);
    comm_complete(scb1,psi_s1); 

    recon(chi_s1,gauge[scb1],this->two_spinor[scb1],&this->simd_cbvol,dag);
    
  }
// Call dperp if Ls!=1.
  if ( dperp && this->Ls > 1 ) { 
    this->Dperp(psi,chi,cb,dag);
  }
  
}
// This routine works out Dperp in DWF
template <class Float>
void bfmbagel1<Float>::Dperp(Fermion_t psi_t, 
			      Fermion_t chi_t, 
			      int cb,
			      int dag)
{

  Float *psi ;
  Float *chi ;
  
  //Interesting dependence on cb because of the BC's
  
  int nspinco = 12;
  int Lsm1 = this->Ls-1;
  int ls_stride = this->simd_cbvol * this->nsimd * nspinco * 2;
  double c;
  int sign;

  if ( dag == 1 ) sign = 1;
  else sign = -1;

  // The bulk forward hopping term
  psi = (Float *) psi_t;  psi+= ls_stride;
  chi = (Float *) chi_t;
  chiralproj_axpby((Fermion_t) chi, (Fermion_t) psi, 2.0 , sign,this->Ls-1);

  // The bulk backward hopping term
  psi = (Float *) psi_t;  
  chi = (Float *) chi_t; chi+= ls_stride;
  chiralproj_axpby((Fermion_t) chi, (Fermion_t) psi, 2.0 ,-sign,this->Ls-1);

  // The Ls-1 upper face forward hopping term
  psi = (Float *) psi_t; 
  chi = (Float *) chi_t; chi+= (this->Ls-1)*ls_stride;
  chiralproj_axpby((Fermion_t) chi, (Fermion_t) psi, -2.0*this->mass , sign,1);

  // The s=0 lower face backward hopping term
  psi = (Float *) psi_t; psi+= (this->Ls-1)*ls_stride;
  chi = (Float *) chi_t; 
  chiralproj_axpby((Fermion_t) chi, (Fermion_t) psi, -2.0*this->mass , -sign,1);

}


template <class Float>
void bfmbagel1<Float>::pointers_end(void)
{
  for ( int cb = 0;cb<Ncb;cb++ ) 
    bfm_free(this->shift_table[cb]);
  for(int cb=0;cb<2;cb++) {
    for(int pm=0;pm<2;pm++) {
      for(int mu=0; mu<4;mu++) {
	bfm_free(this->face_table[cb][pm][mu]);
      }
    }
  }

}
template <class Float>
void bfmbagel1<Float>::pointers_init(void)
{
  int mu;
  int cb, pm;
  int shift;
  int x,y,z,t;
  int lx,ly,lz,lt;

  int SizeofTwoSpin;
 
  int bound_index[Ncb][NMinusPlus][ND];
  int local_p,result_cb;
  int local_addr[ND];
  int shift_addr[ND];

  int psite,shift_psite;
  int offset,tab,table_size;
  void *bit_bucket;

  lx = this->simd_latt[0];
  ly = this->simd_latt[1];
  lz = this->simd_latt[2];
  lt = this->simd_latt[3];

  bit_bucket = bfm_alloc(64);

  if ( BGL ) { 
    if ( this->isBoss() ) printf("Configuring for BG/L kernels\n");
  } else { 
    if ( this->isBoss() ) printf("Configuring for QCDOC kernels\n");
  }

  SizeofTwoSpin = sizeof(Float);

  if ( this->isBoss() ) printf("Padded 2-spinor size is %d words, %d bytes\n",PAD_HALF_SPINOR_SIZE,
	PAD_HALF_SPINOR_SIZE*SizeofTwoSpin);

  for ( cb = 0 ; cb<2 ; cb++) {
    //ND * plus_minus * hvol * parities
    table_size = NMinusPlus*(this->simd_cbvol+1)*Nmu;
    this->shift_table[cb]=(integer *)bfm_alloc( table_size*sizeof(integer));
    if( this->shift_table[cb] == 0x0 ) { 
       if ( this->isBoss() ) printf("bfm::pointers_init: this->shift_tables[%d] alloc failed\n",cb);
       exit(-1);
    }
    if ( this->isBoss() ) printf("bfm::pointers_init: this->shift_tables[%d] alloc \n",cb);
  }

  for(cb=0;cb<2;cb++){
    for(pm=0;pm<2;pm++){
      for(mu=0; mu<4;mu++){

	this->face_table[cb][pm][mu]= (integer *)bfm_alloc(this->simd_allbound*sizeof(integer));
	if( this->face_table[cb][pm][mu] == 0x0 ) { 
	  if ( this->isBoss() ) printf("bfm::pointers_init: this->shift_tables[%d] alloc failed\n",cb);
	  exit(-1);
        }
      }
    }
  }
  
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    bound_index[0][0][mu] = 0;
    bound_index[0][1][mu] = 0;
    bound_index[1][0][mu] = 0;
    bound_index[1][1][mu] = 0;
  }

  /* Now work out the shifts */
  for ( pm = 0; pm<2; pm++ ) {

    /* Shift direction */
    if ( pm == Plus ) { /*forwards or backwards*/
      shift = 1;
    } else { 
      shift = -1;
    }

      /*Loop naively over local lattice*/
    for ( t=0 ; t<lt ; t++ ) {
     local_addr[3] = t;
     for ( z=0 ; z<lz ; z++ ) {
      local_addr[2] = z;
      for ( y=0 ; y<ly ; y++ ) {
       local_addr[1] = y;
       for ( x=0 ; x<lx ; x++ ) {
        local_addr[0] = x;

	/* local_addr contains the coordinates of the point */

	/* get the parity of the point */
        int xx[4] = {x,y,z,t};
        local_p = this->parity(xx);

	/* get the site of the point for that parity */
        psite   = this->psite(local_addr,this->simd_latt);

        for ( mu = 0 ; mu < 4; mu ++) {

	  /* Get the coordinates of the point shifted in direction 
	     +/- mu. (+/- is encoded in shift, depending on where 
	     we are in the +/- loop) */

          shift_addr[0] = x;
          shift_addr[1] = y;
          shift_addr[2] = z;
          shift_addr[3] = t;

          shift_addr[mu] += shift;

	  /* Parity of shifted site is opposite that of the unshifted one */
          result_cb = 1-local_p ;

          /*
	   * Offset to the source site in the 2 spinor array is common
	   */

	  /* Get the interleaved site address of the local point
	     (from which we are shifting */
	  tab = this->interleave_site(pm,mu,psite);

          /*
	   * Get the offset to the destination site in the 2 spinor array
	   * If destination is in interior, trivial.
	   * Also implement periodic wrap if local_comm[mu].
	   */

	  if (((shift_addr[mu] >= 0 ) && (shift_addr[mu]<this->simd_latt[mu]) )
             || this->local_comm[mu] ) {

	    /*Local periodicity, does nothing if interior*/
	    shift_addr[mu]= (shift_addr[mu] + this->simd_latt[mu])%this->simd_latt[mu];
	    shift_psite   = this->psite(shift_addr,this->simd_latt);
	    offset = this->interleave_site(pm,mu,shift_psite) ;
	    this->shift_table[result_cb][tab] = (integer)this->two_spinor[result_cb] +
                   offset * PAD_HALF_SPINOR_SIZE * SizeofTwoSpin;

	  } else { /*non-local and we're on the boundary*/


	    /*Local periodicity, does nothing if interior*/
	    shift_addr[mu]= (shift_addr[mu] + this->simd_latt[mu])%this->simd_latt[mu];
	    shift_psite   = this->psite(shift_addr,this->simd_latt);
	    offset = this->interleave_site(pm,mu,shift_psite) ;

	    /*
	     * The minus face receives something written by the 
	     * plus face. Thus the face table should contain the pointer
	     * to the site we would have sent to in the above case.
	     */
	    this->face_table[result_cb][pm][mu][bound_index[local_p][pm][mu]]
	      = offset;

	    /*
	     * And we use this->face_table[local_p][pm][mu] in conjunction
	     * with the send buffer[pm][mu] and source parity local_p
	     */
	    offset = bound_index[local_p][pm][mu] ;

	    this->shift_table[result_cb][tab] = (integer) this->send_bufs[result_cb][pm][mu]
	      + offset * PAD_HALF_SPINOR_SIZE * SizeofTwoSpin;

	    bound_index[local_p][pm][mu] ++;
	  }

        } /*mu*/
       }/*x*/
      }/*y*/
     }/*z*/
    }/*t*/
  }/*pm*/

  //  int bound_index[Ncb][NMinusPlus][ND];
  for( cb = 0 ; cb < 2 ; cb++ ) {
  for( pm = 0 ; pm < 2 ; pm++ ) {
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    if ( !this->local_comm[mu] ) { 
      if ( bound_index[cb][pm][mu] != this->simd_nbound[mu] ) {
	printf("Boundary size mismatch[%d][%d][mu=%d] : %d != %d \b",
	       cb,pm,mu,bound_index[cb][local_p][mu],this->simd_nbound[mu]);
	exit(-1);
      }
    } else { 
      if ( bound_index[cb][pm][mu] != 0 ) {
	printf("Boundary size mismatch[%d][%d] mu=%d\b",
	       cb,pm,mu);
	exit(-1);
      }
    }
  }
  }
  }

}
template <class Float>
int bfmbagel1<Float>::interleave_site(int mp,int mu, int site)
{
  if ( BGL )
    return(mp+mu*NMinusPlus+site*NMinusPlus*ND);
  else
    return(mu+mp*ND+site*NMinusPlus*ND);
}
template <class Float>
int bfmbagel1<Float>::parity(int xx[4])
{
  int x,y,z,t;
  x = xx[0];
  y = xx[1];
  z = xx[2];
  t = xx[3];
  return((x+y+z+t) % 2);
}
template <class Float>
int bfmbagel1<Float>::psite(int addr[4],int latt[4])
{
  int site;
  site =  addr[0] + latt[0]*(addr[1]
                             + latt[1]*(addr[2]
                                        + latt[2]*addr[3]));
  return(site/2);
}

#ifdef BFM_QDP_BINDING

// This is the base Bagel 2 import. Use QDP to do the shifting, conjugation why not....
template <class Float>
void      bfmbagel1<Float>::importGauge(multi1d<LatticeColorMatrix> &u)
{
  
  if ( u.size() != Nd ) exit(-1);
  for ( int mu=0;mu<Nd;mu++){

    QDPdouble * U_p;

    LatticeColorMatrix uT;
    uT = transpose(u[mu]);
    U_p = (QDPdouble *)&uT.elem(0).elem();
    importGauge(U_p,mu);

  }
}
#endif

template <class Float>
int bfmbagel1<Float>::doubleStored(void) { 
  return 0;
};

template <class Float>
void      bfmbagel1<Float>::importGauge(QDPdouble *gauge, int dir)
{
  int x[4];
  int Ndircoco=36; /*4 directions stored only*/
  int Ncoco=9;

  for ( x[3]=0;x[3]<this->node_latt[3];x[3]++ ) { 
  for ( x[2]=0;x[2]<this->node_latt[2];x[2]++ ) { 
  for ( x[1]=0;x[1]<this->node_latt[1];x[1]++ ) { 
  for ( x[0]=0;x[0]<this->node_latt[0];x[0]++ ) { 
     
      for ( int coco=0;coco<9;coco++ ) { 
      for ( int reim=0;reim<2;reim++ ) { 

        Float * bagel = this->u;
	int bbase = dir*9;
        int cidx = this->chroma_idx(x,reim,coco,Ncoco);
        int bidx = this->bagel_idx(x,reim,coco+bbase,Ndircoco,0);
        bagel[bidx] = gauge[cidx];


      }}

  }}}}
}

template class bfmbagel1<float>;
template class bfmbagel1<double>;


