/****************************************************************************/
/* PAB Dec 2007                                                             */
/*
  // Methods implemented here
  virtual void init(bfmarg &arg) ; 
  virtual void end (void) ;

  virtual integer bagel_idx   (int site[4],int reim,int i, int i_size,int half_cb) = 0;
  virtual integer bagel_idx5d (int site[4],int s,int reim,int i, int i_size,int half_cb) =0 ;
  virtual integer bagel_idx5d (int site,int s,int i_size) =0;

  virtual void dslash(Fermion_t chi, 
		      Fermion_t psi, 
		      int cb,
		      int dag,int dperp=1) ;


  virtual void Dperp(Fermion_t psi, 
		     Fermion_t chi, 
		     int cb,
		     int dag);

  virtual void pointers_init(void) ;
  virtual void pointers_end(void) ; 
*/
/****************************************************************************/

#include <stdlib.h>
#include <stdio.h>

#include "bfm.h"

static double *complex_i_simd;
static double complex_i_simd1[12] __attribute__ ((aligned(32)))  = { 0.0, 1.0, 0.0,-1.0, 0.0, 0.0 };

static double complex_i_simd2[12] __attribute__ ((aligned(32))) = { 0.0, 1.0, 0.0, 1.0, 
                                                                    0.0,-1.0, 0.0,-1.0, 
                                                                    0.0, 0.0, 0.0, 0.0 };

template <class Float>
int bfmbagel2<Float>::simd(void)
{
 return BAGEL_NSIMD;
}

template <class Float>
void bfmbagel2<Float>::init(bfmarg &arg)  
{
  int words;             /* size of the spinor field on the         */
  	  	  	 /* sublattice checkerboard                 */

  int slx;                      /* x-direction size of node sublattice */
  int sly;                      /* y-direction size of node sublattice */
  int slz;                      /* z-direction size of node sublattice */
  int slt;                      /* t-direction size of node sublattice */
  int i;
  int mu;

/*--------------------------------------------------------------------------*/
/* Set sublattice direction sizes                                           */
/*--------------------------------------------------------------------------*/

  * ((bfmarg *)this) = arg;


  /*-------------------------------------------------------------
   * Introduce 5d precon ; use ncb to refer to 4d checkerboards
   *-------------------------------------------------------------
   */
  if ( this->precon_5d && (this->Ls==1) ) { 
    if ( this->isBoss() ) printf("bfm<Float>::init precon_5d & Ls=1 logic bomb\n");
    exit(-1);
  }
  if ( this->precon_5d ) this->ncb=1;
  else this->ncb=2;
  if ( this->precon_5d ) this->cbLs=this->Ls/2;
  else this->cbLs=this->Ls;

  this->node_cbvol = this->node_latt[0]
                               *this->node_latt[1]
                               *this->node_latt[2]
                               *this->node_latt[3]
                               /this->ncb;

  /***************************************************/
  /*This should be selected depending on architecture*/
  if ( this->isBoss() ) printf("nsimd =%d BAGEL2_NSIMD=%d\n",simd(),BAGEL_NSIMD);
  this->nsimd = this->simd();
  this->simd_grid[0] = 1;
  this->simd_grid[1] = 1;
  this->simd_grid[2] = 1;
  this->simd_grid[3] = this->nsimd;
  if (this->nsimd == 1 )      complex_i_simd = complex_i_simd1;
  else if (this->nsimd == 2 ) complex_i_simd = complex_i_simd2;
  else exit(0);
  // SPU case??
  /***************************************************/

  this->simd_latt[0] = slx = this->node_latt[0]/this->simd_grid[0];
  this->simd_latt[1] = sly = this->node_latt[1]/this->simd_grid[1];
  this->simd_latt[2] = slz = this->node_latt[2]/this->simd_grid[2];
  this->simd_latt[3] = slt = this->node_latt[3]/this->simd_grid[3];

  this->base_parity =(this->ncoor[0]*this->node_latt[0] 
              + this->ncoor[1]*this->node_latt[1]
              + this->ncoor[2]*this->node_latt[2]
              + this->ncoor[3]*this->node_latt[3])&0x1;

/*--------------------------------------------------------------------------*/
/* Set periodic wrap back or not                                            */
/*--------------------------------------------------------------------------*/
  this->local_comm[0] = arg.local_comm[0];
  this->local_comm[1] = arg.local_comm[1];
  this->local_comm[2] = arg.local_comm[2];
  this->local_comm[3] = arg.local_comm[3];

/*-----------------------------------------------------------------------*/
/* compute the subgrd volume of each chkbd ... at least two local dims   */
/* must be even for this code to be correct.                             */
/*-----------------------------------------------------------------------*/
  this->simd_cbvol = (slx * sly * slz * slt)/this->ncb;
  
  this->simd_nbound[0] = (sly * slz * slt)/this->ncb; 
  this->simd_nbound[1] = (slx * slz * slt)/this->ncb;
  this->simd_nbound[2] = (slx * sly * slt)/this->ncb;
  this->simd_nbound[3] = (slx * sly * slz)/this->ncb;

  this->simd_allbound  = this->simd_nbound[0]
                 + this->simd_nbound[1]
                 + this->simd_nbound[2]
                 + this->simd_nbound[3];

  if ( this->simd_nbound[0] * slx * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even x logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[1] * sly * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even y logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[2] * slz * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even z logic bomb\n");
    exit(-1);
  }
  if ( this->simd_nbound[3] * slt * this->ncb != (slx*sly*slz*slt) ) {
    if ( this->isBoss() ) printf("bfm::init Even t logic bomb\n");
    exit(-1);
  }

  /*------------------------------------------------------------------------*/
  /* Check shape                                                            */
  /*------------------------------------------------------------------------*/
  if ( (slx&1) || (slz&0x1) || (slt&0x1) ) {
    if ( this->isBoss() ) printf("Bagel is refusing to run as a sub latt is odd\n");
    exit(-1);
  }


/*--------------------------------------------------------------------------*/
/* Reserve memory for 1 spatial site of two spinor                          */
/*--------------------------------------------------------------------------*/
  words = this->cbLs * HALF_SPINOR_SIZE * this->nsimd * 10;

  this->two_spinor = (Float *)bfm_alloc(words*sizeof(Float));
  if(this->two_spinor == 0){
    if ( this->isBoss() ) printf("bfm::two_spinor allocate\n");
    exit(-1);
  }
  bzero((void*)this->two_spinor, words*sizeof(Float));

/*--------------------------------------------------------------------------*/
/* Reserve memory for gauge field                                           */
/*--------------------------------------------------------------------------*/
  words = GAUGE_SIZE * this->nsimd * this->simd_cbvol *2*2;

  this->u = (Float *)bfm_alloc(words*sizeof(Float));
  if(this->u == 0){
    if ( this->isBoss() ) printf("bfm::u allocate\n");
    exit(-1);
  }

/*----------------------------------------------------------------------*/
/* Build the pointer table                                              */
/*----------------------------------------------------------------------*/
  this->pointers_init();
  
/*----------------------------------------------------------------------*/
/* Initialise the comms                                                 */
/*----------------------------------------------------------------------*/
  this->comm_init();

}

template <class Float>
void bfmbagel2<Float>::end (void)
{
  this->comm_end();
  this->pointers_end();
  bfm_free(this->two_spinor);
  bfm_free(this->u);
}

template <class Float>
integer bfmbagel2<Float>::bagel_idx5d (int site,int s,int nspinco) 
{
  if ( this->precon_5d ) s = s/2;
  return s*nspinco*this->simd()*2 + site*nspinco*this->simd()*2*this->cbLs;
}
template <class Float>
integer bfmbagel2<Float>::bagel_idx5d (int x[4],int s,int reim,int i, int i_size, int half_cb)
{
  if ( this->precon_5d ) s = s/2;
  return this->bagel_idx(x,reim,i+s*i_size,i_size*this->cbLs,half_cb);
}
template <class Float>
integer bfmbagel2<Float>::bagel_idx (int x[4],int reim,int i, int i_size, int half_cb)
{
  int gx[4];
  int mu;

  for (mu=0;mu<4;mu++ ) gx[mu] = x[mu]/this->simd_latt[mu];

  integer simd_el = gx[0]+this->simd_grid[0]*( gx[1]
                         +this->simd_grid[1]*( gx[2]
                         +this->simd_grid[2]*  gx[3] ));
  int scoor[4]; // SIMD coordinate

  scoor[0] = x[0] % this->simd_latt[0]; /*Work out our location within the SIMD vol*/
  scoor[1] = x[1] % this->simd_latt[1];
  scoor[2] = x[2] % this->simd_latt[2];
  scoor[3] = x[3] % this->simd_latt[3];

  int scb   = this->parity(scoor);          /*Work out local checkerboard of site*/
  int ssite = this->psite(scoor,this->simd_latt); /*Work out checkerboarded site number*/

  // Real/Imaginary innermost
  int idx = reim + simd_el*2
                 + i*simd()*2
      + ssite*i_size*simd()*2;

// Bagel only deals with half checkerboard fermion vectors
// to allow for red black solution
  if ( !half_cb ) idx+= scb*this->simd()*2*i_size*this->simd_cbvol;

  return idx;
}

/*
 * Implement all the platform specific routines for BG/L,P,Q
 * These are:
 *  bagel_idx -- for mapping the layout in import/export routines
 *  dslash    -- calls the BG assembler
 *  Mooee     -- preconditioning is aware of the data layout 
 *  MooeeInv  -- preconditioning is aware of the data layout 
 *  comm_init,comm_end,comm
 */

static double two_mass_simd1[4] __attribute__ ((aligned(32)));
static double two_mass_simd2[8] __attribute__ ((aligned(32)));


// Aim for CELL support at some time in the future.
// Similar trick expected if I get around to SSE/single
static double complex_i_spu[8] 
  __attribute__ ((aligned(32))) = { 0.0, 0.0, 0.0, 0.0,
                                    1.0, 1.0, 1.0, 1.0 } ;

template <class Float>
void bfmbagel2<Float>::dslash(Fermion_t psi, // Input
              	    Fermion_t chi, // Result
                    int cb,
                    int dag,int dperp)
{

  int lcb = (cb + this->base_parity) & 1;

  this->comm_start(lcb,psi,dag);

//  getarg(Chi_p);         /*Pointer to result spinor          */
//  getarg(Psi_p);         /*Base pointer to input spinor      */
//  getarg(U_p);           /*Pointer to gauge field            */
//  getarg(length);        /*volume of 4d sites in checkerboard*/
//  getarg(Ls);            /*Two spinor                        */
//  getarg(tab);           /*neighbour site table              */
//  getarg(twospinor_base);/*Two spinor (one site * Ls * 10 directions storage)*/
//  getarg(Complex_i);     /*Pointer to complex i              */

  Float *gauge_par;
  /*Gauge args*/
  Float *utmp = (Float *)this->u;
  if ( this->precon_5d ) { 
    gauge_par = utmp;
  } else if ( cb == 1 ) { 
    gauge_par = utmp+ (2*GAUGE_SIZE*this->simd_cbvol*this->simd()); // Double store the gauge field
  } else { 
    gauge_par = utmp;
  }

  /*Turn global checkerboard into the checkerboard within this node*/
  cb = lcb;

  if ( dperp && this->precon_5d) { 
    this->Dperp(psi,chi,cb,dag);
  } else { 
    this->axpby(chi,chi,chi,0.0,0.0); // Zero
  }
  this->comm_complete(lcb,psi);

  integer args[8];
  integer arg_p = (integer)args; 
  args[0] = (integer)chi;
  args[1] = (integer)psi;
  args[2] = (integer)gauge_par;
  args[3] = (integer)this->simd_cbvol;
  args[4] = (integer)this->cbLs;
  if ( this->precon_5d ) {
    args[5] = (integer)this->shift_table[0];
  } else { 
    args[5] = (integer)this->shift_table[cb];
  }
  args[6] = (integer)this->two_spinor;
  args[7] = (integer)complex_i_simd;

  int single =  1;
  if ( sizeof(Float) == sizeof(double) ) single = 0;

  if ( dperp && this->precon_5d) { 
    if ( dag ) {
      if ( single)      vmx_dwf_dag_addto_single(arg_p);
      else              vmx_dwf_dag_addto(arg_p);
    } else {
      if ( single )     vmx_dwf_addto_single(arg_p);
      else              vmx_dwf_addto(arg_p);
    }
  } else { 
    if ( dag ) {
      if (single)   vmx_dwf_dag_addto_single(arg_p);
      else          vmx_dwf_dag_addto(arg_p);
    } else {
      if ( single ) vmx_dwf_addto_single(arg_p);
      else          vmx_dwf_addto(arg_p);
    }
  }

}




// This routine works out Dperp in DWF
template <class Float>
void bfmbagel2<Float>::Dperp(Fermion_t psi, 
		   Fermion_t chi, 
		   int cb,
		   int dag)
{
  //Temporary debug hack 
  two_mass_simd1[0] = two_mass_simd1[1] = 2.0;
  two_mass_simd1[2] = two_mass_simd1[3] =-2.0*this->mass ;

  two_mass_simd2[0] = two_mass_simd2[1] = 2.0;
  two_mass_simd2[2] = two_mass_simd2[3] = 2.0;
  two_mass_simd2[4] = two_mass_simd2[5] =-2.0*this->mass ;
  two_mass_simd2[6] = two_mass_simd2[7] =-2.0*this->mass ;

  integer args[8];
  integer arg_p = (integer)args; 

  args[0] = (integer) psi;
  args[1] = (integer) chi;
  args[2] = (integer) this->simd_cbvol;
  args[3] = (integer) this->cbLs;
  args[4] = (integer) this->cb_table[cb];
  args[5] = (integer) complex_i_simd;

  if(this->simd() == 1 ) args[6] = (integer) two_mass_simd1;
  if(this->simd() == 2 ) args[6] = (integer) two_mass_simd2;

  int single = 1;
  if (sizeof(Float)==sizeof(double) ) single = 0; 
  if ( dag ) {
    if (single ) vmx_dperp_dag_single(arg_p);
    else         vmx_dperp_dag(arg_p);
  } else { 
    if (single ) vmx_dperp_single(arg_p);
    else         vmx_dperp(arg_p);
  }

}



template <class Float>
void bfmbagel2<Float>::pointers_end(void)
{
  for ( int cb = 0;cb<Ncb;cb++ ) {
    bfm_free(this->shift_table[cb]);
    bfm_free(this->cb_table[cb]);
  }
  for(int cb=0;cb<2;cb++) {
    for(int pm=0;pm<2;pm++) {
      for(int mu=0; mu<4;mu++) {
	bfm_free(this->face_table[cb][pm][mu]);
      }
    }
  }
}


/*Layout is:                                     */
/* PSI   [site/Nsimd][Ls][spin][co][reim][Nsimd] */
/* Gauge [site/Nsimd][mu][co][co][reim][Nsimd]   */
/* PSI contains attached face buffers at the back of array */

template <class Float>
void bfmbagel2<Float>::pointers_init(void)
{

  int mu;
  int cb, pm;
  int shift;
  int x,y,z,t;
  int lx,ly,lz,lt;
 
  int bound_index[Ncb][NMinusPlus][ND];
  int local_p;
  int local_addr[ND];
  int shift_addr[ND];

  int cbsite,shift_cbsite;
  int offset,tab,table_size;

  lx = this->simd_latt[0];
  ly = this->simd_latt[1];
  lz = this->simd_latt[2];
  lt = this->simd_latt[3];

  for ( cb = 0 ; cb<2 ; cb++) {
    table_size = NMinusPlus*(this->simd_cbvol+1)*Nmu;
    this->shift_table[cb]=(integer *)bfm_alloc( table_size*sizeof(integer));
  }
  for ( cb = 0 ; cb<2 ; cb++) {
    table_size = (this->simd_cbvol+8);
    this->cb_table[cb]=(integer *)bfm_alloc( table_size*sizeof(integer));
  }

  for(cb=0;cb<2;cb++){
    for(pm=0;pm<2;pm++){
      for(mu=0; mu<4;mu++){

	this->face_table[cb][pm][mu]= (integer *)
	     bfm_alloc(this->simd_allbound*sizeof(integer));

      }
    }
  }
  
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    bound_index[0][0][mu] = 0;
    bound_index[0][1][mu] = 0;
    bound_index[1][0][mu] = 0;
    bound_index[1][1][mu] = 0;
  }

  /*
   * This bit points to the first element after the end of the
   * body for comm_offset[Minus][0] 
   */
  this->comm_offset[Minus][0]  = this->simd_cbvol; 

  /* Now we do the rest of the comms offsets packing densely.
     nbound[mu-1] sites after the previous */
  this->comm_offset[Minus][1]  = this->comm_offset[Minus][0] + this->simd_nbound[0];
  this->comm_offset[Minus][2]  = this->comm_offset[Minus][1] + this->simd_nbound[1];
  this->comm_offset[Minus][3]  = this->comm_offset[Minus][2] + this->simd_nbound[2];

  /* Now the PLUS directions. They are the same as the minus directions 
     pushed by allbound */
  this->comm_offset[Plus][ 0 ] = this->comm_offset[Minus][0] + this->simd_allbound;
  this->comm_offset[Plus][ 1 ] = this->comm_offset[Minus][1] + this->simd_allbound;
  this->comm_offset[Plus][ 2 ] = this->comm_offset[Minus][2] + this->simd_allbound;
  this->comm_offset[Plus][ 3 ] = this->comm_offset[Minus][3] + this->simd_allbound;


  /* Now work out the shifts */
  for ( pm = 0; pm<2; pm++ ) {

    /* Shift direction */
    if ( pm == Plus ) { /*forwards or backwards*/
      shift = 1;
    } else { 
      shift = -1;
    }

      /*Loop naively over local lattice*/
    for ( t=0 ; t<lt ; t++ ) {
     local_addr[3] = t;
     for ( z=0 ; z<lz ; z++ ) {
      local_addr[2] = z;
      for ( y=0 ; y<ly ; y++ ) {
       local_addr[1] = y;
       for ( x=0 ; x<lx ; x++ ) {
        local_addr[0] = x;

	/* local_addr contains the coordinates of the point */
	/* we get the parity of the point                   */
        local_p = this->parity(local_addr);
	/* get the cbsite of the point for within parity */
        cbsite   = this->psite(local_addr,this->simd_latt);

	// Used for DWF Dperp
	this->cb_table[0][cbsite] = (x+y+z+t)&0x1;
	this->cb_table[1][cbsite] = 1-this->cb_table[0][cbsite];

        for ( mu = 0 ; mu < 4; mu ++) {

	  /* Get the coordinates of the point shifted in direction 
	     +/- mu. (+/- is encoded in shift, depending on where 
	     we are in the +/- loop) */

          shift_addr[0] = x;
          shift_addr[1] = y;
          shift_addr[2] = z;
          shift_addr[3] = t;

          shift_addr[mu] += shift;

	  /* Get the interleaved site address of the local point
	     (to which we are shifting data for dslash stencil*/
	  tab = this->interleave_site(pm,mu,cbsite);

          /*we're on the boundary*/
	  int on_boundary = 0;
	  if ( (shift_addr[mu] < 0) || (shift_addr[mu]>=this->simd_latt[mu] ) ) {
	    on_boundary	= 1;
          }
          /*
	   * Get the offset to the neighbour site in the spinor with periodicity
	   */
          shift_addr[mu]= (shift_addr[mu] + this->simd_latt[mu])%this->simd_latt[mu];
          shift_cbsite   = psite(shift_addr,this->simd_latt);

          if ( on_boundary == 0 ) { 

	    this->shift_table[local_p][tab] = shift_cbsite;

	  } else {

	    /*
	     * The face_table contains a pointer to the
	     * neighbour site, to be gathered into the comms buffers.
             * local_p is the RESULT checkerboard/parity
	     * This should be sent in the OPPOSITE (1-pm) direction.
	     * Corresponding received data (same if loopback) 
	     * is stored the comm_offset[pm]
	     */
	    int bdx = bound_index[local_p][pm][mu];
	    this->face_table[local_p][pm][mu][bdx] = shift_cbsite;

	    /*
	     * And we use face_table[local_p][pm][mu] in conjunction
	     * with the send buffer[pm][mu] and source parity local_p
	     */
            if ( (!this->local_comm[mu]) || (this->simd_grid[mu]>1) ) {
	      this->shift_table[local_p][tab] = this->comm_offset[pm][mu] + bdx;
            } else { 
	      this->shift_table[local_p][tab] = shift_cbsite;
            }
	    /*
	     *   X X X A
	     * 
	     *   face_table[plus] = 4
	     *   shift_table[4,plus] = this->comm_offset[plus]
	     *
	     *   gather (face_table) -> buffer
             *   This should be copied to -> this->comm_offset[minus]
	     *   Reading the corresponding minus face from shift_table pulls A in the hopping term
	     */
	    bound_index[local_p][pm][mu] ++;
	  }


        } /*mu*/
       }/*x*/
      }/*y*/
     }/*z*/
    }/*t*/
  }/*pm*/

  //  int bound_index[Ncb][NMinusPlus][ND];
  for( cb = 0 ; cb < 2 ; cb++ ) {
  for( pm = 0 ; pm < 2 ; pm++ ) {
  for( mu = 0 ; mu < Nmu ; mu++ ) {
    if ( (!this->precon_5d) || cb == 0 ) {
      if ( bound_index[cb][pm][mu] != this->simd_nbound[mu] ) {
	printf("Boundary size mismatch[%d][%d][mu=%d] : %d != %d \n",
	       cb,pm,mu,bound_index[cb][pm][mu],this->simd_nbound[mu]);
//	exit(-1);
      }
    }
  }
  }
  }

  return;
}

template <class Float>
int bfmbagel2<Float>::interleave_site(int mp,int mu, int site)
{
  return(mp+mu*NMinusPlus+site*NMinusPlus*ND);
}

//
// To switch to 5d preconditioning, I think I just need to
// make parity always return zero, and make psite not divide by two.
// introduce a flat precon_5d in the bfmarg class
//

template <class Float>
int bfmbagel2<Float>::parity(int x[4])
{
  if ( this->precon_5d ) return 0;
  else return((x[0]+x[1]+x[2]+x[3]) % 2);
}

template <class Float>
int bfmbagel2<Float>::psite(int addr[4],int latt[4])
{
  int site;
  site =  addr[0] + latt[0]*(addr[1]
                             + latt[1]*(addr[2]
                                        + latt[2]*addr[3]));
  if ( this->precon_5d ) return site;
  else return(site/2);
}

template <class Float>
void bfmbagel2<Float>::Dperp_cpp(Fermion_t psi_t, 
			   Fermion_t chi_t, 
			   int cb,
			   int dag)
{

  Float *psi = (Float *) psi_t;
  Float *chi = (Float *) chi_t;
  
  //Interesting dependence on cb because of the BC's
  
  int nspinco = 12;
  int x[4];

  for ( int s=0;s<this->Ls;s++ ) { 
  for ( x[3]=0;x[3]<this->node_latt[3];x[3]++ ) { 
  for ( x[2]=0;x[2]<this->node_latt[2];x[2]++ ) { 
  for ( x[1]=0;x[1]<this->node_latt[1];x[1]++ ) { 
  for ( x[0]=0;x[0]<this->node_latt[0];x[0]++ ) { 

    int p = (s+x[3]+x[2]+x[1]+x[0])& 0x1;

    if ( p == cb ) { 

      Float c_p,  c_m ;
      Float *p_m,*p_p,*ch;
      int s_p,s_m;

      int site = this->psite(x,this->simd_latt);

      c_m = c_p = 2.0;

      if ( s== 0    ) c_m = -2*this->mass; // mf boundary wrap
      if ( s== this->Ls-1 ) c_p = -2*this->mass;

      s_p = (s+1)%this->Ls;
      s_m = (s+this->Ls-1)%this->Ls;

      p_m = &psi[this->bagel_idx5d(site,s_m,nspinco)];
      p_p = &psi[this->bagel_idx5d(site,s_p,nspinco)];
      ch  = &chi[this->bagel_idx5d(site,s,nspinco)];

      for(int spinco=0;spinco<nspinco/2;spinco++){
      for(int simd_el=0;simd_el<this->nsimd;simd_el++){
      for(int reim=0;reim<2;reim++){

        int idx = reim + simd_el*2 + spinco*this->nsimd*2;
        int hdx = reim + simd_el*2 + (spinco+6)*this->nsimd*2;

        if ( dag ) {
          ch[idx]+= c_p*p_p[idx]; // 1+gamma_5
          ch[hdx]+= c_m*p_m[hdx]; // 1-gamma_5
        }else{
          ch[idx]+= c_m*p_m[idx]; // 1+gamma_5
          ch[hdx]+= c_p*p_p[hdx]; // 1-gamma_5
        }

      }}}


    }
  }}}}}

}


static int silent;

template <class Float>
void bfmbagel2<Float>::gather_face(integer *sites,
			     Float *psi,
			     integer comm_off, 
			     Float *comm_buf,
			     integer nface,
			     int permute, 
			     int cbLs
			     )
{

  if ( ! permute ) { 
    if ( sizeof(Float) == sizeof(double) ) { 
      vmx_gather(sites, (double *)psi, comm_off, (double *) comm_buf, nface, cbLs);
    } else { 
      vmx_gather_single(sites, (float *)psi, comm_off, (float *) comm_buf, nface, cbLs);
    }
    return;
  }

  int nspinco = 12;

  for(int bdx=0;bdx<nface;bdx++){
    integer site  = sites[bdx];
    integer bsite = comm_off + bdx;

    for(int s=0;s<this->cbLs;s++){
    for(int spinco=0;spinco<nspinco;spinco++){
    for(int simd_el=0;simd_el<this->simd();simd_el++){
    for(int reim=0;reim<2;reim++){
    
      // Real/Imaginary innermost
      int idx = reim + simd_el*2
                     + spinco*this->simd()*2
                     + s*nspinco*this->simd()*2
                     + site*nspinco*this->simd()*2*this->cbLs;

      int jsimd= simd_el;
      if ( permute ) jsimd = (1-jsimd);

      int jdx = reim + jsimd*2
                     + spinco*this->simd()*2
                     + s*nspinco*this->simd()*2
                     + bsite*nspinco*this->simd()*2*this->cbLs;

       /*
	*Pulls the face into a comms buffer
        *(in fake this lives in) the tail of the array
        */
       comm_buf[jdx] = psi[idx];

    }}}}
  }
}
template <class Float>
void bfmbagel2<Float>::gather_proj(integer *sites,
			     Float     *psi,
			     integer comm_off, 
			     Float  *comm_buf,
			     integer nface,
			     int permute, 
			     int L,
			     int pm,
			     int mu)
{
  integer args[8];
  args[0] = (integer) sites;
  args[1] = (integer) psi;
  args[2] = (integer) comm_off;
  args[3] = (integer) comm_buf;
  args[4] = (integer) nface;
  args[5] = (integer) L;
  args[6] = pm+2*mu;
  args[7] = (integer)complex_i_simd;

  integer arg_p = (integer) args;
  if ( sizeof(Float) == sizeof(double ) ) vmx_gather_proj(arg_p);
  else                                    vmx_gather_proj_single(arg_p);
}
		   
template class bfmbagel2<float>;
template class bfmbagel2<double>;
