#include "bfm.h"
#include <stdio.h>
#include <stdlib.h>



#include "bfmcommqmpbagel2.h"
#include <qmp.h>

#define STRIDED_RECEIVE

/*
 * QMP comms implementation
 * If ( local_comm[mu]== 1 ) these routines do not get called
 * If ( local_comm[mu]== 0 ) these routines get called but loop back
 *
 * For now a brain dead implementation with no spin projection.
 */

#define recv_multi_handle multi_handle[0]
#define send_multi_handle multi_handle[1]

template <class Float> 
void commQMPbagel2<Float>::recv_end(void)
{
  if ( num_op>0 ) {
    QMP_free_msghandle(recv_multi_handle);
    //    QMP_free_msghandle(all_handles);
  }
  for(int mu=0; mu < ND; mu++) { 
    if (!this->local_comm[mu]) { 
      QMP_free_msgmem(recv_ops_msgmem_t[2*mu]);
      QMP_free_msgmem(recv_ops_msgmem_t[2*mu+1]);
#ifndef STRIDED_RECEIVE      
      bfm_free(recvbufs[2*mu]);
      bfm_free(recvbufs[2*mu+1]);
#endif
    }
  }
}
template <class Float> 
void commQMPbagel2<Float>::recv_init(Fermion_t psi_t)
{
  num_op = 0;
  Float *psi = (Float *) psi_t;  
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++ ) {
      if( !this->local_comm[mu] ) {

	int idx;
	Float * ptr;
	
	// Allocate send/receive buffers
	// Declare send & msgmem
	int blocks = this->simd_nbound[mu] * this->cbLs;
	int block  = HALF_SPINOR_SIZE * this->simd() *sizeof(Float);
        int stride =      SPINOR_SIZE * this->simd() *sizeof(Float);
	int words  = this->simd_nbound[mu] * HALF_SPINOR_SIZE * this->simd() * this->cbLs;

	int delta;

	if ( pm == Plus ) delta = 1;
	else delta = -1;

	idx=2*mu+pm; 

	//
	// We pull the "Plus" face_table and send in the Minus direction
	// This reflects an oddity of the 
#ifdef STRIDED_RECEIVE
	ptr= &psi[this->comm_offset[pm][mu]*SPINOR_SIZE*this->simd()*this->cbLs];
	recv_ops_msgmem_t[idx]      = QMP_declare_strided_msgmem((void *)ptr,block,blocks,stride);
#else
	ptr = recvbufs[idx] = (Float *)bfm_alloc(words*sizeof(Float)); 
	recv_ops_msgmem_t[idx]  = QMP_declare_msgmem((void *)ptr,words*sizeof(Float));
	bzero(ptr,words*sizeof(Float));
#endif 
	recv_handles[num_op++]  = QMP_declare_receive_relative(recv_ops_msgmem_t[idx], mu, delta, 0);

      } // if ( !local_comm )
    }      
    } // for(mu=... )
    
    if( num_op > 0 ) { 
      recv_multi_handle = QMP_declare_multiple(recv_handles, num_op);
      //      all_handles       = QMP_declare_multiple(multi_handle, 2);
    }

  return;

}
template <class Float> 
void commQMPbagel2<Float>::comm_init(void)
{
                                                                            
  num_op = 0;
    
  for(int mu=0;mu<4;mu++){
    for(int pm=0;pm<2;pm++ ) {
      if( !this->local_comm[mu] ) {

	int words, idx;
	Float * ptr;
	
	// Allocate send/receive buffers
	// Declare send & msgmem
	
	words = this->simd_nbound[mu] * HALF_SPINOR_SIZE * this->simd() * this->cbLs;
	int delta;
	if ( pm == Plus ) delta = 1;
	else delta = -1;

	idx=2*mu+pm; 
	//
	// We pull the "Plus" face_table and send in the Minus direction
	// This reflects an oddity of the 
	ptr= sendbufs[idx] = (Float *)bfm_alloc(words*sizeof(Float)); 
	send_ops_msgmem_t[idx] = QMP_declare_msgmem((void *)ptr,words*sizeof(Float));
	send_handles[num_op++] = QMP_declare_send_relative(send_ops_msgmem_t[idx], mu, -delta, 0);

      } // if ( !local_comm )
    }      
    } // for(mu=... )
    
    if( num_op > 0 ) { 
      send_multi_handle = QMP_declare_multiple(send_handles, num_op);
    }

#ifndef STRIDED_RECEIVE   
  recv_init((Fermion_t)NULL);
#endif
  return;
}

template <class Float> 
void commQMPbagel2<Float>::comm_end(void)
{
  /*-----------------------------------------------------------------------* 
   * I need to free the multiple msghandles                                *
   * The individual handles (tmp_handles) used should have been freed      *
   * when the multiple was created and according to the spec I don't need  *
   * to free them                                                          *
   *---------------------------------------------------------------------- */
  if ( num_op>0 ) QMP_free_msghandle(send_multi_handle);
  /*---------------------------------------------------------------------*
   * I need to free the associated msgmem_t-s here.                      *
   *---------------------------------------------------------------------*/
  for(int mu=0; mu < ND; mu++) { 
    if (!this->local_comm[mu]) { 
      QMP_free_msgmem(send_ops_msgmem_t[2*mu]);
      QMP_free_msgmem(send_ops_msgmem_t[2*mu+1]);
      bfm_free(sendbufs[2*mu]);
      bfm_free(sendbufs[2*mu+1]);
    }
    
  }
#ifndef STRIDED_RECEIVE   
  recv_end();
#endif

  return;
}

template <class Float> 
void commQMPbagel2<Float>::comm  (int result_cb, Fermion_t psi,int dag)  
{
  comm_start(result_cb,psi,dag);
  comm_complete(result_cb,psi);
}


template <class Float> 
void commQMPbagel2<Float>::comm_start  (int result_cb, Fermion_t psi,int dag)  
{
  // gather the faces.
  gather (result_cb,psi,dag);
  comm_qmp_start(psi);
  return;
}
template <class Float> 
void commQMPbagel2<Float>::comm_qmp_start(Fermion_t psi) 
{
#ifdef STRIDED_RECEIVE
  recv_init(psi);
#endif
  if ( num_op > 0 ) {
    QMP_start(send_multi_handle);
    QMP_start(recv_multi_handle);
  }
}
template <class Float> 
void commQMPbagel2<Float>::comm_qmp_complete(void) 
{
  if ( num_op > 0 ) { 

    
    if( QMP_wait(recv_multi_handle) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }
    
    if( QMP_wait(send_multi_handle) != QMP_SUCCESS ) { 
      if ( isBoss() ) printf("QMP_wait returned unsuccessful value: \n");
      exit(-1);
    }

  }
}

template <class Float> 
void commQMPbagel2<Float>::comm_complete  (int result_cb, Fermion_t psi)  
{
  comm_qmp_complete() ;
#ifdef STRIDED_RECEIVE
  recv_end();
#else
  scatter (result_cb, psi);
#endif
  return;
}

template <class Float> 
bool commQMPbagel2<Float>::isBoss(void)
{
  return QMP_is_primary_node();
}

template <class Float> 
void commQMPbagel2<Float>::gather(int result_cb, Fermion_t psi,int dag)
{
  int nspinco = 12;
  int cb = 1-result_cb;

/**********************************************************
 * Gather the faces into the communications buffers
 **********************************************************
 *
 * Fake implementation - just gather the upper face directly 
 *                       into the (lower) receive buffer
 *
 * -> Gathers buffers at comm offset[pm][mu] from
 *    face_table[result_cb][pm][mu]
 **********************************************************
 */

  for(int mu=0;mu<ND;mu++){
  if ( (!this->local_comm[mu]) || (this->simd_grid[mu]>1) ) {

    int permute = 0;  
    if ( this->simd_grid[mu] > 1 ) {
      /*The permute here is tricky. 
       *If we need to permute and non-local we should
       *do a half-word extract & merge for the positive/negative face.
       */
      if ( !this->local_comm[mu] ) { 
	if ( isBoss() ) { 
	  printf("Oops need to implement face skimming code for SIMD words\n");
	}
	exit(-1);
      }
      permute = 1;
    }

    if ( this->simd_grid[mu] > 2 ) exit(0); // Not supported yet
    if ( this->simd()        > 2 ) exit(0); // Not supported yet

    int ftcb ;
    if ( this->precon_5d ) ftcb = 0;
    else ftcb = result_cb;
    for(int pm=0;pm<2;pm++) {

      int sgn =pm;
      if ( dag ) sgn = 1-pm;
      gather_proj(this->face_table[ftcb][pm][mu],
		  (Float *)psi,
	           0,sendbufs[2*mu+pm],
		  this->simd_nbound[mu],
		  permute,this->cbLs,sgn,mu);

    }


  }}

}


template <class Float> 
void commQMPbagel2<Float>::scatter(int result_cb, Fermion_t psi_t)
{
  Float *psi = (Float *) psi_t;
  for(int mu=0;mu<ND;mu++){
    for(int pm=0;pm<2;pm++){
    if ( (!this->local_comm[mu]) || (this->simd_grid[mu]>1) ) {
      Float * optr = &psi[this->comm_offset[pm][mu]*SPINOR_SIZE*this->simd()*this->cbLs];
      Float * iptr = recvbufs[mu*2+pm];

      if ( sizeof(Float) == sizeof(double) ) { 
	vmx_scatter((double *)iptr,(double *)optr,this->simd_nbound[mu],this->cbLs);
      } else { 
	vmx_scatter_single((float *)iptr,(float *)optr,this->simd_nbound[mu],this->cbLs);
      }
    } 
  }
  }
}

template <class Float> 
void commQMPbagel2<Float>::comm_gsum(double &val) 
{
  QMP_sum_double_array(&val,1);
}

template class commQMPbagel2 <float>;
template class commQMPbagel2 <double>;

