/*
 *
 *  Copyright UKQCD Collaboration, March 2007.
 *  Written by Peter Boyle.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "processor.h"
extern struct processor *PROC;

#include "registers.h"

void dwf_dperp( char *);

/*Options flags*/
int human = 0;
char name[80];
int dagger = 0;

void do_writehint(int,int);
#define do_WH(A,B) ;

Datum FourSpinType=Double;
Datum GaugeType=Double;
int addto = 0;
char procname[80]="UNSPECIFIED";

int main ( int argc, char *argv[])
{
  struct processor *myproc;
  int arg;
  char *c;

  name[0] = '\0';
  /*Process command line*/
  while ( ( arg = getopt(argc,argv,"aRn:P:ds")) != EOF){
    switch(arg){
    case 'R': human = 1; break;
    case 'n': if (strlen(optarg)<30) strcpy(name,optarg); break;
    case 'P': if (strlen(optarg)<30) strcpy(procname,optarg); break;
    case 'd': dagger = 1; break;
    case 'a': addto = 1; break;
    case 's': FourSpinType= GaugeType = Single; break;
    default: fprintf(stderr,"Usage: %s -[Rd] -n routine_name\n",argv[0]); 
             fprintf(stderr,"\tR -> human readable .S format \n"); 
             fprintf(stderr,"\td -> dagger decompose\n"); 
             exit (1); break;
    }
  }

  /*Control output according to user set up args*/
  set_processor_optarg(procname);
  set_human_readable(human);
  setup_cmadds();

  /*Write the naive asm code*/
  dwf_dperp(name);

  /*Filter through an virtual out of order processor */
  schedule_for_proc();

  /*Dump the resulting code*/
  dump_instruction_queue();

  return(0);
}

void dwf_dperp( char *name)
{
  /*
   * This marks the argument registers as defined by ABI as off limits
   * to us until they are freed by "getarg()";
   */
  int dum = defargcount(1);

  /*Handles for the labels point*/
  int branchsite;
  int branchmu;
  int retno;

  /*-------------------------------------------------------------------------------
   * registers used 
   *-------------------------------------------------------------------------------
   */
  reg_array_2d(Chi ,Cregs,4,3);// 2 spionr      - 6 regs 
  reg_array_2d(Psi ,Cregs,4,3); /*Output 4-spinor - 12 regs*/
  alreg(creg,Cregs);
  alreg(coeffm,Cregs);
  alreg(coeffp,Cregs);

  offset_3d(CHIIMM,FourSpinType,2,3,2*nsimd());
  offset_1d(COEFF,Double,4*nsimd());

  int t;

  /*
   * Integer registers
   */
  alreg(Psi_p,Iregs);   /*Pointer to current cpt in/out PSI field      */
  alreg(Chi_p,Iregs);   /*Pointer to current cpt in/out PSI field      */
  alreg(length,Iregs);  /*number of sites*/
  alreg(Complex_i,Iregs);  /*Point to (0,1)x Nsimd*/
  alreg(Ls,Iregs);        
  alreg(tab,Iregs);        

  alreg(site,Iregs);  /*number of sites*/
  alreg(sloop,Iregs);        
  alreg(args,Iregs);        
  alreg(tabent,Iregs);        
  alreg(s,Iregs);        
  alreg(s_m,Iregs);        
  alreg(s_p,Iregs);        
  alreg(tmp,Iregs);        
  alreg(ptr,Iregs);        
  alreg(coeff_p,Iregs);        
  alreg(pref,Iregs);
  alreg(pref1,Iregs);

  /*Useful integer immediate constants, in units of Fsize*/
  def_off( ZERO_IMM,Byte,0);
  def_off( CHI_IMM, FourSpinType,12*nsimd());
  def_off( PSI_IMM, FourSpinType,24*nsimd());

  int Isize      = def_offset(PROC->I_size,Byte,"Isize");
  int negone     = def_offset(-1,Byte,"negone");
  int plusone    = def_offset(1,Byte,"plusone");

  /*
   * Local variables for C++ code
   */
  int i,sp,co,j,k,nxt,ri;

  /*********************************************************************
   * Start of the "pseudo assembler proper.
   *********************************************************************/
  make_inst(DIRECTIVE,Enter_Routine,name);
  grab_stack(0);
  save_regs();

  /*
   * Define our arguments 
   */
  getarg(args); /*Pointer to arg list*/

  pragma(DCBT_SPACE,0);

  queue_iload(Chi_p, ZERO_IMM,args);  queue_load_addr(args,Isize,args);  
  queue_iload(Psi_p, ZERO_IMM,args);  queue_load_addr(args,Isize,args);  
  queue_iload(length,ZERO_IMM,args);  queue_load_addr(args,Isize,args);  
  queue_iload(Ls,    ZERO_IMM,args);  queue_load_addr(args,Isize,args);  
  queue_iload(tab,   ZERO_IMM,args);  queue_load_addr(args,Isize,args);  
  queue_iload(Complex_i,ZERO_IMM,args);    queue_load_addr(args,Isize,args);  
  queue_iload(coeff_p,ZERO_IMM,args);   

  /*Bugger 12 int regs for offsets as indexed loads only*/
  for ( i =0; i<6; i++ ) { 
    need_constant(i*2*SizeofDatum(FourSpinType)*nsimd());
  }
  int   two = args;
  int   negtwomass = coeff_p;

  queue_load_addr(two,COEFF[0],coeff_p);  
  queue_load_addr(negtwomass,COEFF[nsimd()*2],coeff_p);  

  /*
   * Site loop
   */
  queue_iadd_imm(pref,Chi_p,ZERO_IMM);    
  queue_iadd_imm(pref,pref,PSI_IMM);    
  queue_iadd_imm(pref,pref,PSI_IMM);    

  queue_iadd_imm(pref1,Psi_p,ZERO_IMM);    
  queue_iadd_imm(pref1,pref1,PSI_IMM);    

  queue_iload_imm(site,ZERO_IMM);
  retno = get_target_label(); /*Branch to exit if length <1*/
  check_iterations(length,retno); 
  branchsite = start_loop(length);

    /*
     * Here parity is NOT easy to compute as a function of site use a lookup table
     */
    queue_iload(tabent,ZERO_IMM,tab);
    queue_load_addr(tab,Isize,tab);

    /*Fifth dimension loop*/
    queue_iadd_imm(sloop,Ls,ZERO_IMM);
    queue_iload_imm(s,ZERO_IMM);


    int branchls = start_loop(sloop);

      do_prefetch(pref,0);
      do_prefetch(pref,1);
      if ( FourSpinType ==Double ) 
	do_prefetch(pref,2);
      queue_iadd_imm(pref,pref,CHI_IMM);

      queue_iadd_imm(s_p,s,plusone); 
      queue_isub(tmp,s_p,Ls);

      int tmp_ge_lab = get_target_label();
      int tmp_le_lab = get_target_label();

      conditional_branch_cmpzero ( BRANCH_LT, tmp, tmp_ge_lab  ) ;

        make_inst(IALUPIPE,IOR,s_p,tmp,tmp);   
        make_inst(IALUPIPE,IOR,ptr,negtwomass,negtwomass);   
	conditional_branch_cmpzero( BRANCH_GT,tabent,tmp_le_lab);

      make_inst(DIRECTIVE,CmovTarget,tmp_ge_lab);
      make_inst(IALUPIPE,IOR,ptr,two,two);   
      make_inst(DIRECTIVE,CmovTarget,tmp_le_lab);
      //Need to load constant depending on whether peri-wrapped
      complex_load(coeffp,ZERO_IMM,ptr,Double);

      if ( addto ) { 
	do_prefetch(pref1,0);
	do_prefetch(pref1,1);
	if ( FourSpinType ==Double ) 
	  do_prefetch(pref1,2);
	queue_iadd_imm(pref1,pref1,CHI_IMM);
      }

      /*checkerboarded site hopping arithmetic
       *If result_cb == cb4d then neighbours on source_cb are s-1,s
       *If result_cb != cb4d then neighbours on source_cb are s,s+1
       */
      queue_iadd_imm(s_m,s,negone); 
      /*    if ( sm < 0 ) { */
      int sm_lt_lab = get_target_label();
      int sm_ge_lab = get_target_label();

      conditional_branch_cmpzero(BRANCH_GE,s_m,sm_lt_lab);
        make_inst(IALUPIPE,IOR,ptr,negtwomass,negtwomass);   
        queue_iadd(s_m,s_m,Ls);
	conditional_branch_cmpzero(BRANCH_LE,tabent,sm_ge_lab);

      /* } else { */
      make_inst(DIRECTIVE,CmovTarget,sm_lt_lab);
        make_inst(IALUPIPE,IOR,ptr,two,two);    
      /* } */
      make_inst(DIRECTIVE,CmovTarget,sm_ge_lab);
      complex_load(coeffm,ZERO_IMM,ptr,Double);

      do_prefetch(pref,0);
      do_prefetch(pref,1);
      if ( FourSpinType ==Double ) 
	do_prefetch(pref,2);
      queue_iadd_imm(pref,pref,CHI_IMM);

// If 4d parity IS the result CB, s_p is same site on other CB      
      queue_cmovle(tabent,s,s_p); 
      queue_cmovgt(tabent,s,s_m);

      /* Compute addresses */
      if ( addto ) {
	do_prefetch(pref1,0);
	do_prefetch(pref1,1);
	if ( FourSpinType ==Double ) 
	  do_prefetch(pref1,2);
	queue_iadd_imm(pref1,pref1,CHI_IMM);
      }
      //This work is loop invariant
      queue_imul(tmp,site,Ls);    
      queue_imul_imm(tmp,tmp,PSI_IMM);    //tmp is offset to s=0 for site

      queue_iadd(ptr,tmp,Chi_p);    

      queue_imul_imm(s_p,s_p,PSI_IMM);    
      queue_iadd(s_p,s_p,ptr);    


      if ( !dagger ) queue_iadd_imm(s_p,s_p,CHI_IMM);
      for(int sp=0;sp<2;sp++){
	for(int co=0;co<3;co++){
	  if ( !dagger )
	    complex_load(Chi[sp+2][co],CHIIMM[sp][co][0],s_p,FourSpinType);
	  else
	    complex_load(Chi[sp][co],CHIIMM[sp][co][0],s_p,FourSpinType);
	}
      }

      queue_imul_imm(s_m,s_m,PSI_IMM);    
      queue_iadd(s_m,s_m,ptr);    

      if ( dagger ) queue_iadd_imm(s_m,s_m,CHI_IMM);
      for(int sp=0;sp<2;sp++){
	for(int co=0;co<3;co++){
	  if ( dagger )
	    complex_load(Chi[sp+2][co],CHIIMM[sp][co][0],s_m,FourSpinType);
	  else
	    complex_load(Chi[sp][co],CHIIMM[sp][co][0],s_m,FourSpinType);
	}
      }

      queue_iadd(ptr,tmp,Psi_p);    // Offset from the base pointers
      queue_imul_imm(tmp,s,PSI_IMM);
      queue_iadd(ptr,tmp,ptr);  
      queue_iadd_imm(tmp,ptr,ZERO_IMM);   
      for(int osp=0;osp<2;osp++){
      for(int isp=0;isp<2;isp++){
	int sp=osp*2+isp;
	for(int co=0;co<3;co++){
	  if ( addto ) {
	    complex_load (Psi[sp][co],CHIIMM[isp][co][0],ptr,FourSpinType);
	    if ( dagger ) { 
	      if ( sp>=2 ) { 
		complex_real_madd  (Psi[sp][co],coeffm,Chi[sp][co]);
	      } else { 
		complex_real_madd  (Psi[sp][co],coeffp,Chi[sp][co]);
	      }
	    } else { 
	      if ( sp>=2 ) { 
		complex_real_madd  (Psi[sp][co],coeffp,Chi[sp][co]);
	      } else { 
		complex_real_madd  (Psi[sp][co],coeffm,Chi[sp][co]);
	      }
	    }
	  } else {
	    if ( dagger ) { 
	      if ( sp>=2 ) { 
		complex_real_mul  (Psi[sp][co],coeffm,Chi[sp][co]);
	      } else { 
		complex_real_mul  (Psi[sp][co],coeffp,Chi[sp][co]);
	      }
	    } else { 
	      if ( sp>=2 ) { 
		complex_real_mul  (Psi[sp][co],coeffp,Chi[sp][co]);
	      } else { 
		complex_real_mul  (Psi[sp][co],coeffm,Chi[sp][co]);
	      }
	    }
	  }
	}
      }
      queue_iadd_imm(ptr,ptr,CHI_IMM);
      }

      for(int osp=0;osp<2;osp++){
      for(int isp=0;isp<2;isp++){
	int sp=osp*2+isp;
	for(int co=0;co<3;co++){
	  complex_store(Psi[sp][co],CHIIMM[isp][co][0],tmp,FourSpinType);
	}
      }
      queue_iadd_imm(tmp,tmp,CHI_IMM);
      }

      queue_iadd_imm(s,s,plusone);
    stop_loop(branchls,sloop);

    queue_iadd_imm(site,site,plusone);
  /*
   * TERMINATION point of the loop
   */
  stop_loop(branchsite,length);
  make_inst(DIRECTIVE,Target,retno);

  /*
   * EPILOGUE
   */
  restore_regs();
  free_stack();
  make_inst(DIRECTIVE,Exit_Routine,name);

  return;
}
  
