#include <bfm.h>
#include <stdio.h>
#include <stdlib.h>

typedef double Float;
class singletest : public bfm {
public:

  void unitGauge(void)
  {
    int vol = node_latt[0]
            * node_latt[1]
            * node_latt[2]
            * node_latt[3];
    int words = 18 * vol;

    int bytes = sizeof(QDPdouble)*words;

    QDPdouble *gauge = (QDPdouble *)malloc(bytes);
    bzero(gauge,bytes);

    // No easy way to support bagel-2      
    // without CHROMA/QDP as I used QDP to do the cshift & adjoint
    // for double stored links. just use unit gauge.
    for(int v=0;v<vol;v++ ) {
      gauge[v*18]    = 1.0;
      gauge[v*18+8]  = 1.0;
      gauge[v*18+16] = 1.0;
    }
    int ndir=4;
    if ( doubleStored() ) { 
      ndir = 8;
    }
    for(int mu=0;mu<ndir;mu++) {
      importGauge(gauge, mu); 
    }
  }
  void randFermion(Fermion_t x_t){

    int vol = node_latt[0]
            * node_latt[1]
            * node_latt[2]
            * node_latt[3];
    int words = 24 * vol;
    int bytes = sizeof(QDPdouble)*words;

    QDPdouble *psi = (QDPdouble *)malloc(bytes);

    for(int w=0;w<words;w++ ) {
      psi[w] = drand48();
    }

    importFermion(psi, x_t, 0);

  };

  void debugFermion(Fermion_t x_t){

    int vol = node_latt[0]
            * node_latt[1]
            * node_latt[2]
            * node_latt[3];
    int words = 24 * vol;
    int bytes = sizeof(QDPdouble)*words;

    QDPdouble *psi = (QDPdouble *)malloc(bytes);

    for(int v=0;v<vol;v++ ) {
    for(int w=0;w<24;w++ ) {
      if ( w < 6 )
	psi[w+v*24] = v*100+w;
      else 
	psi[w+v*24] = w;
    }
    }
    importFermion(psi, x_t, 0);

  };


void dump(Fermion_t psi, char *file)
{
  int vol = node_latt[0]
          * node_latt[1]
          * node_latt[2]
          * node_latt[3];

  int words = 24 * vol;

  int bytes = sizeof(QDPdouble)*words;

  QDPdouble * psi_p = (QDPdouble *)malloc(bytes);
  bzero(psi_p, bytes);

  exportFermion(psi_p,psi,0);

  FILE * fp = fopen(file,"w");
  for(int s=0;s<vol;s++){
    for(int sc=0;sc<24;sc++){
      fprintf(fp,"%le\n",(double)*psi_p++);
    }
  }
  fclose(fp);
};
void load(Fermion_t psi, char *file)
{
  int vol = node_latt[0]
          * node_latt[1]
          * node_latt[2]
          * node_latt[3];

  int words = 24 * vol;

  int bytes = sizeof(QDPdouble)*words;

  QDPdouble * psi_p = (QDPdouble *)malloc(bytes);
  QDPdouble * tmp = psi_p;
  FILE * fp = fopen(file,"r");
    for(int s=0;s<vol;s++){
    for(int sc=0;sc<24;sc++){
      double d;
      fscanf(fp,"%le",&d);
      *tmp++ = d;
    }
    }
  fclose(fp);
  importFermion(psi_p,psi,0);
};
};

int main (int argc,char **argv )
{

  int generate = 0;
  if ( argc > 1 ) generate = 1;

  int lx = 4;
  int ly = 4;
  int lz = 4;
  int lt = 4;

  int nrow[4];
  nrow[0] = lx;
  nrow[1] = ly;
  nrow[2] = lz;
  nrow[3] = lt;

  bfmarg dwfa;

  singletest    dwf;

  dwfa.node_latt[0]  = lx;
  dwfa.node_latt[1]  = ly;
  dwfa.node_latt[2]  = lz;
  dwfa.node_latt[3]  = lt;

  dwfa.local_comm[0]  = 0;
  dwfa.local_comm[1]  = 0;
  dwfa.local_comm[2]  = 0;
  dwfa.local_comm[3]  = 0;

  dwfa.Ls = 1;
  dwfa.mass = 0.0;
  dwfa.Csw  = 0.0;

  printf("Initialising bfm operator\n");
  printf("drand48 seed = 0\n");
  srand48(0);

  dwf.init(dwfa);

  Fermion_t psi_h = dwf.allocFermion();
  Fermion_t chi_h = dwf.allocFermion();
  Fermion_t check = dwf.allocFermion();
  Fermion_t diff  = dwf.allocFermion();
  dwf.randFermion(psi_h);
  //  dwf.debugFermion(psi_h);
  dwf.unitGauge();

  
  // Naive Dslash

  // cb is cb of result, 1-cb is cb of input field
  char * files[4] = { "cb0dag0.dat",
                      "cb0dag1.dat",
                      "cb1dag0.dat",
                      "cb1dag1.dat" };
  int idx=0;
  for(int cb=0;cb<2;cb++){
    
    /*Import this checkerboard of QDP fields to bagel*/
    printf("Importing psi field cb %d\n",cb);

    // Fill the other checkerboard.
    for(int dag=0;dag<2;dag++){

      printf("Checking cb=%d dag=%d\n",cb,dag); 

      dwf.dslash(psi_h,
                 chi_h, 
		 cb,dag);

      if ( generate) {
	dwf.dump(chi_h,files[idx++]);
      } else {
	dwf.load(check,files[idx++]);
	dwf.axpy(diff,check,chi_h,-1.0);
	double delta = dwf.norm(diff);
	printf("Norm of difference is %le\n",delta);
      }
      printf("Done\n"); 

    }
  }

}







